@@ -1,30 +1,36 @@
+
 from lxml.etree import ElementBase
 from lxml.html import fromstring, html_parser, tostring, fragments_fromstring
 from lxml.builder import ElementMaker
+
+from cssselect import HTMLTranslator
+
 from uuid import uuid4
-import collections
 
 
 E = ElementMaker(makeelement=html_parser.makeelement)
+css_to_xpath = HTMLTranslator().css_to_xpath
+
+
+def _matches(node, selector, data=None, index=0):
+    sel = _select_all(node, selector, data=data, index=index, prefix='self::')
+    sel = [e for e in sel if e is not None and e is not False]
+    return bool(len(sel))
 
 
 def _select(node, selector, data=None, index=0):
-    sel = _selectAll(node, selector, data=data, index=index)[:1]
+    sel = _select_all(node, selector, data=data, index=index)[:1]
     return None if len(sel) == 0 else sel[0]
 
 
-def _selectAll(node, selector, data=None, index=0):
+def _select_all(node, selector, data=None, index=0, prefix='descendant::'):
     if callable(selector):
         ans = selector(node, data, index)
         return ans if isinstance(ans, list) else [ans]
     elif isinstance(selector, ElementBase):
         return [selector]
 
-    nodes = node.cssselect(selector)
-    if node in nodes:
-        nodes.remove(node)
-
-    return nodes
+    return node.xpath(css_to_xpath(selector, prefix=prefix))
 
 
 def _fake_node(d, ds):
@@ -35,9 +41,9 @@ def _fake_node(d, ds):
 
 class Group(object):
     """Contains nodes, and a pointer
-    to the parentNode for this group"""
+    to the parent_node for this group"""
     def __init__(self, parent, nodes=None):
-        self.parentNode = parent
+        self.parent_node = parent
         self.nodes = nodes if nodes is not None else list()
 
     def __len__(self):
@@ -107,13 +113,13 @@ class EnterSelection(BaseSelection):
 
         for group in self:
             update = group.updates
-            subgroup = Group(group.parentNode)
+            subgroup = Group(group.parent_node)
             sel.append(subgroup)
             for i, node in enumerate(group):
                 if node is not None:
                     data = ds.get(node, None)
 
-                    subnode = _select(group.parentNode, selector, data, i)
+                    subnode = _select(group.parent_node, selector, data, i)
                     subgroup.append(subnode)
 
                     if subnode is not None:
@@ -131,7 +137,7 @@ class Selection(BaseSelection):
     def __init__(self, root):
         super(Selection, self).__init__(root)
 
-    def selectAll(self, selector):
+    def select_all(self, selector):
         sel = Selection(self.root)
         ds = self.dataset
 
@@ -140,7 +146,7 @@ class Selection(BaseSelection):
                 if node is not None:
                     sel.append(Group(
                         node,
-                        _selectAll(node, selector, ds.get(node, None))
+                        _select_all(node, selector, ds.get(node, None))
                     ))
 
         return sel
@@ -149,7 +155,7 @@ class Selection(BaseSelection):
         sel = Selection(self.root)
 
         for group in self:
-            subgroup = Group(group.parentNode)
+            subgroup = Group(group.parent_node)
             sel.append(subgroup)
             for i, node in enumerate(group):
                 if node is not None:
@@ -165,6 +171,18 @@ class Selection(BaseSelection):
 
         return sel
 
+    def filter(self, selector):
+        sel = Selection(self.root)
+
+        for group in self:
+            subgroup = Group(group.parent_node)
+            sel.append(subgroup)
+            for i, node in enumerate(group):
+                d = self.dataset.get(node, None)
+                if node is not None and _matches(node, selector, d, i):
+                    subgroup.append(node)
+        return sel
+
     def remove(self):
         def _remove(node, d, i):
             node.getparent().remove(node)
@@ -246,12 +264,13 @@ class Selection(BaseSelection):
         return self
 
     def data(self, data=None):
-        if data is None:
-            return self.dataset.get(self.node(), None)
 
-        def bind(group):
+        if data is None:
+            return [self.dataset.get(node, None) for node in self[0]]
+
+        def bind(group, values):
             n = len(group)
-            m = len(data)
+            m = len(values)
             n0 = min(n, m)
             updates = [None] * m
             enters = [None] * m
@@ -259,7 +278,7 @@ class Selection(BaseSelection):
 
             for i in range(n0):
                 node = group[i]
-                node_data = data[i]
+                node_data = values[i]
                 if node is not None:
                     self.dataset[node] = node_data
                     updates[i] = node
@@ -267,15 +286,15 @@ class Selection(BaseSelection):
                     enters[i] = _fake_node(node_data, self.dataset)
 
             for i in range(n0, m):
-                enters[i] = _fake_node(data[i], self.dataset)
+                enters[i] = _fake_node(values[i], self.dataset)
 
             for i in range(m, n):
                 exits[i] = group[i]
 
-            exit.append(Group(group.parentNode, exits))
+            exit.append(Group(group.parent_node, exits))
 
-            ugroup = Group(group.parentNode, updates)
-            egroup = Group(group.parentNode, enters)
+            ugroup = Group(group.parent_node, updates)
+            egroup = Group(group.parent_node, enters)
             egroup.updates = ugroup
 
             update.append(ugroup)
@@ -289,7 +308,9 @@ class Selection(BaseSelection):
         update.enter = lambda: enter
         update.exit = lambda: exit
 
-        [bind(group) for group in self]
+        for i, group in enumerate(self):
+            bind(group, data if not callable(data) else data(
+                 group, self.dataset.get(group.parent_node, None), i))
 
         return update
 
@@ -318,9 +339,9 @@ class P3(object):
         sel.append(subgroup)
         return sel
 
-    def selectAll(self, selector):
+    def select_all(self, selector):
         sel = Selection(self)
         subgroup = Group(self.document)
-        subgroup.append(_selectAll(self.document, selector))
+        subgroup.append(_select_all(self.document, selector))
         sel.append(subgroup)
         return sel
